"""
@des: 这个文件仅仅作为views，接口调用访问，归因内部不用这个文件
@des: 调用方法
from tfduck.common.defines import BMOBJ, Et
tdq = ThinkDataQuery("http://xxxxxxx:xxx/querySql", token="xxxxxxxxxx")
try:
    sql = ''' select * from v_user_7 limit 100  '''
    local_file = tdq.get_data_csv({}, sql, block_size=50000)
    df = pandas.read_csv(local_file, header=0)
finally:
    BMOBJ.remove_file(local_file)

版本记录:
pyhive=0.6.2 
requests=2.23.0 2.27.1
"""
import requests
import pandas
import json
import time
import os
import uuid
import urllib3
from tfduck.common.defines import BMOBJ, Et
# from django.conf import settings
from concurrent.futures import ThreadPoolExecutor, wait, ALL_COMPLETED, FIRST_COMPLETED
from tfduck.tga.base_tga import BaseTga

class ThinkDataQuery(BaseTga):
    """
    @des: thinkdata openapi查询基础类----这个只能再thinkdata内网执行
    """

    def __init__(self, query_uri, token, hive_conn_info=['host', 0]):
        """
        @des:初始化类
        """
        self.query_uri = query_uri  # "http://47.90.251.214:8992/querySql"
        self.token = token
        self.hive_conn_info = hive_conn_info

    def gen_local_unique_file(self, ext="csv"):
        """
        @des:生成本地文件唯一路径
        """
        # media_root = settings.MEDIA_ROOT
        media_root = "/mydata/media"
        base_dir = os.path.join(media_root, "docs")
        if not os.path.exists(base_dir):
            os.makedirs(base_dir)
        real_name = "%s%s.%s" % (BMOBJ.get_unique_id2(), BMOBJ.get_unique_id2(), ext)
        file_path = os.path.join(base_dir, real_name)
        return file_path

    def g_to_csv_notmp(self, filepath, df, index=True, compression=None, mode='w', header=True):
        """
        @des: pandas生成csv文件---用于追加文件，不能用临时文件
        compression: 压缩格式 ‘gzip’, ‘bz2’, ‘zip’, ‘xz’. 
        """
        tmp_filepath = filepath
        if index is None:  # 不保存行索引
            if compression is None:  # 不压缩
                df.to_csv(tmp_filepath, index=None, mode=mode, header=header)
            else:
                df.to_csv(tmp_filepath, index=None,
                          compression=compression, mode=mode, header=header)
        else:
            if compression is None:  # 不压缩
                df.to_csv(tmp_filepath, mode=mode, header=header)
            else:
                df.to_csv(tmp_filepath, compression=compression,
                          mode=mode, header=header)
        return True

    def get_data_csv_i(self, ctx, unique_path, sql, block_size=100000, print_size=100000, read_timeout=600, upcount=None):
        """
        @des:从thinkdata的openapi获取数据----流式，为了节省内存---配合下面的getquerycsv
        """
        session = requests.session()
        post_data = {'token': self.token, 'sql': sql}
        #
        unique_path = self.gen_local_unique_file()
        #
        BMOBJ.log_error("in query")
        #
        r = session.post(self.query_uri, data=post_data, stream=True,
                         verify=False, timeout=(30, read_timeout))
        datas = []
        i = 0    # 循环引用计数
        icount = 0  # 数据的数量
        cols = []  # 表头
        try:
            # iter_lines iter_content, chunk_size字节, 下面取100M
            for row in r.iter_lines(chunk_size=1024*1024*100):
                if not row:
                    continue
                data = None
                if i == 0:  # 处理header
                    data = json.loads(row)
                    if(data["return_code"] == 0):
                        cols = data["data"]["headers"]
                        df = pandas.DataFrame(data=[], columns=cols)  # 保存表头
                        self.g_to_csv_notmp(unique_path, df, index=None)
                        data = None
                    else:
                        BMOBJ.log_error("sql error:", data)
                        # BMOBJ.log_error(sql)
                        datas = []
                        break  # 表示查询出错，没有消息
                else:
                    if row.strip() not in [b"", ""]:
                        data = json.loads(row)
                if data is not None:
                    datas.append(data)
                i += 1
                if len(datas) == block_size:  # 1000000条保存一次
                    df = pandas.DataFrame(data=datas, columns=cols)  # 保存表头
                    self.g_to_csv_notmp(unique_path, df, index=None,
                                        mode='a', header=False)  # 追加保存
                    icount += block_size
                    datas = []
                if i % print_size == 0:
                    BMOBJ.clog(ctx, i)
            BMOBJ.clog(ctx, f"total: {i}")
            if len(datas) > 0:  # 保存最后收尾的
                df = pandas.DataFrame(data=datas, columns=cols)  # 保存表头
                self.g_to_csv_notmp(unique_path, df, index=None,
                                    mode='a', header=False)  # 追加保存
                icount += len(datas)
                datas = []
        except Exception as e:
            BMOBJ.clog(ctx, "get data error", str(e))
            if upcount is not None:
                if i < upcount:  # 看是否达到可以接受的数量，否则重新查询
                    raise e
            else:
                raise e
        finally:
            try:
                r.close()
            except:
                pass
            try:
                session.close()
            except:
                pass
        return unique_path

    def get_data_csv(self, ctx, sql, block_size=100000, print_size=100000, read_timeout=600, upcount=None, retry_count=2):
        """
        @des:从thinkdata的openapi获取数据----流式，为了节省内存---配合下面的getquerycsv
        """
        # unique_path = "./test.csv"
        unique_path = self.gen_local_unique_file()
        gol_e = None
        for i in range(retry_count):
            try:
                result = self.get_data_csv_i(
                    ctx, unique_path, sql, block_size, print_size, read_timeout, upcount)
                return result
            except Exception as e:
                gol_e = e
                BMOBJ.remove_file(unique_path)
                BMOBJ.remove_folder(unique_path)
                if str(e).find("Read timed out") != -1 or str(e).find("Connection broken") != -1:
                    BMOBJ.clog(ctx, f'retry {i}')
                    continue
                else:
                    break
        if gol_e is not None:
            raise gol_e
    
    def get_data_raw_pyhive(self, ctx, sql, block_size=100000, fetch_size=10000, retry_count=2, read_timeout=300, upcount=None, print_size=100000):
        """
        @des: 接口装饰器--修改为get_data_csv，防止全面修改代码
        """
        result = self.get_data_csv(ctx, sql, block_size, print_size, read_timeout, upcount, retry_count)
        return result
        

    def get_data_raw_pyhive_bck(self, ctx, sql, block_size=100000, fetch_size=10000, retry_count=2, read_timeout=300, upcount=None, print_size=100000):
        '''
        @des:presto直连方式读取-----重试的方式----当get_data_csv接口出问题，则启用这个接口
        tobj = ThinkDataQuery("http://queryhost:port/querySql", "查询token",
                          ["presto直连的host", 直连的port])  
        sql = """select * from v_event_7 where "$part_date"='2022-02-24' limit 100 """
        unique_path = tobj.get_data_raw_pyhive({}, sql)
        '''
        # unique_path = "./test.csv"
        unique_path = self.gen_local_unique_file()
        gol_e = None
        for i in range(retry_count):
            try:
                result = self.get_data_raw_pyhive_i(
                    ctx, unique_path, sql, block_size, fetch_size, read_timeout, upcount, print_size)
                return result
            except Exception as e:
                gol_e = e
                BMOBJ.remove_file(unique_path)
                BMOBJ.remove_folder(unique_path)
                if str(e).find("Read timed out") != -1:
                    BMOBJ.clog(ctx, f'retry {i}')
                    continue
                else:
                    break
        if gol_e is not None:
            raise gol_e

    def get_data_raw_pyhive_i(self, ctx, unique_path, sql, block_size=100000, fetch_size=10000, read_timeout=300, upcount=None, print_size=100000):
        '''
        @des: 内部调用
        '''
        from pyhive import presto
        #
        # unique_path = self.gen_local_unique_file()
        # unique_path = "./test.csv"
        #
        BMOBJ.log_error("in query")
        session = requests.session()
        #
        datas = []
        i = 0    # 循环引用计数
        icount = 0  # 数据的数量
        cols = []  # 表头
        try:
            conn = presto.connect(host=self.hive_conn_info[0],
                                  port=int(self.hive_conn_info[1]),
                                  username='ta', catalog='hive',
                                  schema='ta',
                                  requests_session = session,
                                  # 这里stream为true和false没有关系，fetchmany每次都会通过request_session传nexturl重新get获取数据
                                  # 参考pyhive/presto.py的_fetch_more，每次fetchmany其实是多次fetchone
                                  requests_kwargs={"timeout": (
                                      30, read_timeout), "stream": True, "verify": False}
                                  )
            cursor = conn.cursor()
            cursor.execute(sql)
            BMOBJ.clog(ctx, "文件大小")
            if 1:
                cols = [item[0] for item in cursor.description]
                # print(cols)
                df = pandas.DataFrame(data=[], columns=cols)  # 保存表头
                self.g_to_csv_notmp(unique_path, df, index=None)
            # for row in cursor.fetchall():
            rows = []
            # def yx_fetch_many():
            #     for kkk in range(5):  # 最多重试五次
            #         try:
            #             myres = cursor.fetchmany(fetch_size)
            #             return myres
            #         except Exception as e:
            #             if str(e).find("Read timed out")!=-1:
            #                 continue
            #             else:
            #                 raise e

            def yx_fetch_many():
                myres = cursor.fetchmany(fetch_size)
                return myres
            rows = yx_fetch_many()
            while rows:
                for row in rows:
                    if not row:
                        continue
                    datas.append(row)
                    i += 1
                    if len(datas) == block_size:  # 1000000条保存一次
                        df = pandas.DataFrame(data=datas, columns=cols)  # 保存表头
                        self.g_to_csv_notmp(unique_path, df, index=None,
                                            mode='a', header=False)  # 追加保存
                        icount += block_size
                        datas = []
                    if i % print_size == 0:
                        BMOBJ.clog(ctx, i)
                rows = yx_fetch_many()
            #
            BMOBJ.clog(ctx, f"total: {i}")
            if len(datas) > 0:  # 保存最后收尾的
                df = pandas.DataFrame(data=datas, columns=cols)  # 保存表头
                self.g_to_csv_notmp(unique_path, df, index=None,
                                    mode='a', header=False)  # 追加保存
                icount += len(datas)
                datas = []
        except Exception as e:
            # BMOBJ.log_error("get data error", e)
            BMOBJ.clog(ctx, "get data error:", f"{e}")
            if upcount is not None:
                if i < upcount:  # 看是否达到可以接受的数量，否则重新查询
                    raise e
            else:
                raise e
        finally:
            try:
                conn.close()
            except:
                pass  
            try:
                session.close()
            except:
                pass 
        return unique_path



# coding=utf-8
import gzip
import io
import json
import time
import urllib

import requests
import requests.adapters
import six

from tfduck.common.defines import BMOBJ

# class BM(object):
#     def log_error(self, *args, **kwargs):
#         try:
#             if args:
#                 print(args)
#             if kwargs:
#                 print(kwargs)
#         except:
#             pass
# BMOBJ = BM()

# opends_api_url_prefix = 'https://open.bdp.cn'
opends_api_url_prefix = 'http://172.22.129.1:8361' # 李波搭建的bdp传输后台 外网http://47.89.187.14:8361 内网：http://172.22.129.1:8361
# yx_socks5_proxy = {"http": "socks5://127.0.0.1:1080",
#                    "https": "socks5://127.0.0.1:1080"}
yx_socks5_proxy = {
    "http": "http://bdp-proxy:okbdp@161.117.59.90:80",  # "http://47.101.168.174:8899",
    "https": "http://bdp-proxy:okbdp@161.117.59.90:80",  # "http://47.101.168.174:8899"
}
# 走[国内代理]的好处，因为[BDP服务器]在北京，[BDP服务器]本身有超时机制，传输数据超时就触发连接中断错误，导致各种connection abort,ssl oef,enconet等错误
# 数据从[美东]->[国内代理]，因为[国内代理]自己设置的，不存在[美东]->[国内代理]有数据超时的问题
# 如果是[国内代理]-->[BDP服务器]，因为国内传输很快，所以就不会触发BDP的超时机制
use_socks5_proxy = False
need_redirect = False
VERSION = "1.0.3"


class OpenDSException(Exception):
    pass


class OpenDS:
    token = None

    def __init__(self, token=None, domain=None):
        self.url_prefix = opends_api_url_prefix
        requests.packages.urllib3.disable_warnings()
        self.token = token
        if need_redirect:
            self.url_prefix = self.service_uri(domain)

    def public_request(self, url, payload=None, param=None):
        if not payload:
            payload = {}
        if not param:
            param = {}
        param['_t'] = time.time()
        param['access_token'] = self.token
        try_count = 0
        result = {}
        while True:
            try:
                params = u'%s' % urllib.parse.urlencode(param)
                _url = "%s?%s" % (url, params)
                # BMOBJ.log_error("requests default retry count: %s" %
                #                 requests.adapters.DEFAULT_RETRIES)
                if payload:
                    headers = {
                        'Content-type': 'text/html;charset=utf-8',
                        # 'Content-Encoding': 'gzip',
                        "User-Agent": "bdp-sdk-python, {version}".format(version=VERSION)
                    }
                    payload_str = u'%s' % json.dumps(payload)
                    # gzip
                    # s = io.StringIO() #only python2
                    # g = gzip.GzipFile(fileobj=s, mode='w')
                    if 0:
                        s = six.BytesIO()  # python 3
                        payload_str = payload_str.encode("utf8")
                        g = gzip.GzipFile(fileobj=s, mode='w')
                        g.write(payload_str)
                        g.close()
                    else:
                        payload_str = payload_str.encode("utf8") # 不用gzip的方式
                    if use_socks5_proxy:
                        res = requests.post(
                            _url, data=payload_str, headers=headers, verify=False, timeout=(60, 3600), proxies=yx_socks5_proxy).text  # 连接超时60秒，读取超时3600秒
                    else:
                        res = requests.post(
                            _url, data=payload_str, headers=headers, verify=False, timeout=(60, 3600)).text  # 连接超时60秒，读取超时3600秒
                else:
                    headers = {
                        'Content-type': 'text/html;charset=utf-8',
                        # 'Content-Encoding': 'gzip',
                        "User-Agent": "bdp-sdk-python, {version}".format(version=VERSION)
                    }
                    if use_socks5_proxy:
                        res = requests.get(
                            _url, headers=headers, verify=False, timeout=(60, 3600), proxies=yx_socks5_proxy).text  # 连接超时60秒，读取超时3600秒
                    else:
                        res = requests.get(
                            _url, headers=headers, verify=False, timeout=(60, 3600)).text  # 连接超时60秒，读取超时3600秒
                result = json.loads(res)

                break
            except requests.exceptions.ConnectTimeout as e:  # 连接超时，可以重试
                try_count += 1
                BMOBJ.log_error(
                    'can not connect to server, retry ... | reason: %s' % str(e))
                time.sleep(5)
                if try_count == 5:
                    raise e
            except requests.exceptions.ReadTimeout as e1:  # 读取错误，不要重试，否则会出现数据重复或者重复commit的现象
                raise e1
            # ECONNRESET错误不能重试
            # HttpServer默认设置了超时时间为2分钟，当一个请求的处理时间超过2分钟，HttpServer会自动将该请求的socket关闭掉，于是客户端便收到了 ECONNRESET 的错误信息了
            # 修改每次上传的最大条数可以减少这个ECONNRESET的出现问题
            # 例如: (Caused by SSLError(SSLError("bad handshake: SysCallError(104, 'ECONNRESET')",),))
            # except IOError as e3: # 有时候是其他IO错误
            #     try_count += 1
            #     BMOBJ.log_error(
            #         'can not connect to server, retry ... | reason: %s' % str(e3))
            #     time.sleep(5)
            #     if try_count == 5:
            #         raise e3
            except Exception as e2:
                raise e2

        try:
            BMOBJ.log_error(u'api:%s\t request_id:%s' % (
                '/'.join(url.split('/')[-2:]), result.get('request_id', '')))
        except Exception as e:
            BMOBJ.log_error("log_error error")
        if not result:
            error_msg = "\nstatus: {status}\nerror_str: {errstr}\nrequest_id: {req_id}".format(
                status="-1",
                errstr="connection failed",
                req_id=""
            )
            raise OpenDSException(error_msg)
        if result['status'] != '0':

            error_msg = "\nstatus: {status}\nerror_str: {errstr}\nrequest_id: {req_id}".format(
                status=result['status'],
                errstr=result['errstr'],
                req_id=result.get('request_id', '')
            )
            raise OpenDSException(error_msg)
        return result

    def _request(self, url, payload=None, param=None):
        result = self.public_request(url, payload, param)
        return result['result']

    def _request_yx(self, url, payload=None, param=None):
        result = self.public_request(url, payload, param)
        return (result['result'], result.get('request_id', 'no_request_id'))

    def ds_create(self, token, name):
        url = '%s/api/ds/create' % self.url_prefix
        params = {
            'access_token': token,
            'name': name,
            'type': 'opends'
        }
        return self._request(url, param=params)

    def ds_list(self, token):
        url = '%s/api/ds/list' % self.url_prefix
        params = {
            'access_token': token
        }
        return self._request(url, param=params)

    def ds_delete(self, token, ds_id):
        url = '%s/api/ds/delete' % self.url_prefix
        params = {
            'access_token': token,
            'ds_id': ds_id
        }
        return self._request(url, param=params)

    def tb_create(self, token, ds_id, name, schema, uniq_key, title=None):
        url = '%s/api/tb/create' % self.url_prefix
        params = {
            'access_token': token,
        }
        data = {
            'ds_id': ds_id,
            'name': name,
            'schema': schema,
            'uniq_key': uniq_key
        }
        if title:
            data['title'] = title

        return self._request(url, param=params, payload=data)

    def tb_name_modify(self, token, tb_id, alias_name):
        url = '%s/api/tb/modify' % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id,
            'name': alias_name
        }
        return self._request(url, param=params)

    def tb_delete(self, token, tb_id):
        url = '%s/api/tb/delete' % self.url_prefix

        params = {
            'access_token': token,
            'tb_id': tb_id
        }
        return self._request(url, param=params)

    def tb_list(self, token, ds_id):
        url = '%s/api/tb/list' % self.url_prefix
        params = {
            'access_token': token,
            'ds_id': ds_id
        }
        return self._request(url, param=params)

    def tb_info(self, token, tb_id):
        url = '%s/api/tb/info' % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id
        }
        return self._request(url, param=params)

    def tb_clean(self, token, tb_id):
        url = '%s/api/tb/clean' % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id
        }
        return self._request(url, param=params)

    def tb_commit(self, token, tb_id):
        url = '%s/api/tb/commit' % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id,
            'fast': 0
        }
        return self._request(url, param=params)

    def tb_merge(self, token, tb_id):
        url = '%s/api/tb/commit' % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id,
            'fast': 0
        }
        return self._request(url, param=params)

    def tb_update(self, token, tb_ids):
        url = '%s/api/tb/update' % self.url_prefix
        params = {
            'access_token': token,
            'tb_ids': json.dumps(tb_ids)
        }
        return self._request(url, param=params)

    def tb_insert(self, token, tb_id, fields, data):
        url = '%s/api/tb/insert' % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id,
            'fields': json.dumps(fields)
        }
        return self._request(url, param=params, payload=data)

    def tb_insert_by_id(self, token, tb_id, fields, data):
        url = '%s/api/tb/insert' % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id,
            'fields_id': json.dumps(fields)
        }
        return self._request(url, param=params, payload=data)

    def data_insert(self, token, tb_id, fields, data):
        url = '%s/api/data/insert' % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id,
            'fields': json.dumps(fields)
        }
        return self._request(url, param=params, payload=data)

    def data_insert_by_id(self, token, tb_id, fields, data):
        url = '%s/api/data/insert' % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id,
            'fields_id': json.dumps(fields)
        }
        return self._request(url, param=params, payload=data)

    def tb_preview(self, token, tb_id):
        url = '%s/api/tb/preview' % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id,
        }
        return self._request(url, param=params)

    def tb_revert(self, token, tb_id):
        url = '%s/api/tb/revert' % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id,
        }
        return self._request(url, param=params)

    def field_list(self, token, tb_id):
        url = '%s/api/field/list' % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id
        }
        return self._request(url, param=params)

    def field_del(self, token, tb_id, name):
        url = '%s/api/field/delete' % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id,
            'name': name
        }
        return self._request(url, param=params)

    def field_add(self, token, tb_id, name, type, uniq_index, title=None):
        url = '%s/api/field/add' % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id,
            'name': name,
            'type': type,
            'uniq_index': uniq_index
        }
        if title is not None:
            params["title"] = title

        return self._request(url, param=params)

    def field_modify(self, token, tb_id, name, f_type, uniq_index, title=None):
        url = '%s/api/field/modify' % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id,
            'name': name,
            'type': f_type,
            'uniq_index': uniq_index
        }
        if title is not None:
            params["title"] = title
        return self._request(url, param=params)

    def data_update(self, token, tb_id, fields, data):
        url = "%s/api/data/update" % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id,
            'fields': json.dumps(fields)
        }

        return self._request(url, param=params, payload=data)

    def data_update_by_id(self, token, tb_id, fields, data):
        url = "%s/api/data/update" % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id,
            'fields_id': json.dumps(fields)
        }

        return self._request(url, param=params, payload=data)

    def data_delete(self, token, tb_id, fields, data):
        url = "%s/api/data/delete" % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id,
            'fields': json.dumps(fields)
        }

        return self._request(url, param=params, payload=data)

    def data_delete_by_id(self, token, tb_id, fields, data):
        url = "%s/api/data/delete" % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id,
            'fields_id': json.dumps(fields)
        }

        return self._request(url, param=params, payload=data)

    def data_bulkdelete(self, token, tb_id, where):
        url = "%s/api/data/bulkdelete" % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id,
            'where': where
        }
        return self._request(url, param=params)

    def data_bulkdelete_yx(self, token, tb_id, where):
        url = "%s/api/data/bulkdelete" % self.url_prefix
        params = {
            'access_token': token,
            'tb_id': tb_id,
            'where': where
        }
        return self._request_yx(url, param=params)

    def service_uri(self, domain):
        url = "%s/api/service/uri" % self.url_prefix
        params = {
            'domain': domain,
        }
        return self._request(url, param=params)

    def service_error(self):
        url = "%s/api/service/error" % self.url_prefix
        return self._request(url)



# -*- coding: utf-8 -*-
"""
本文件夹内执行测试
"""
from train_sql_ltv import TrainFeatureSql
# from predict_sql_ltv import TrainFeatureSql  # 预测的用法和以前一样

def test():
    class GConf(object):
        def __init__(self):
            self.part_date_start = "2023-01-15"
            self.part_date_end = "2023-01-17"
            # 注册后0-14天的所有ltv作为标签
            self.label_durs = [int(86400*0), int(86400*14)]
            self.level_durs = [86400*0.25, 86400*0.5,
                               86400, 86400*2, 86400*3]  # 由小到大排列
            self.tga_user_table = "v_user_30"
            self.tga_event_table = "v_event_30"

    gconf = GConf()

    need_events = (
        'new_device',
        'new_session',
        'g_push',
        'sdk_close_session'
    )
    need_event_attrs = (
        "#lib",
        "pn",
        "#device_model",
        "#screen_width",
        "#screen_height",
        "#country_code",
        "$part_event",
        "#user_id"
    )
    feature_names = {
        'b': '打开游戏次数',
        'c': '本地推送打开游戏次数'
    }
    addon_attrs = [
        "#device_model",
        "#screen_width",
        "#screen_height"
    ]
    sub_sql_fs = {
        'b':
        # 打开游戏次数
        """
                                --%(real_des)s
                                sum(
                                    if(
                                    "$part_event"='new_session' and rt_dur<%(level_dur)s,
                                    1, 
                                    0)
                                ) as %(sub_field_name)s
                                """, 'c':
        # 本地推送打开游戏次数
        """
                                --%(real_des)s
                                sum(
                                    if(
                                    "$part_event"='g_push' and  rt_dur<%(level_dur)s,
                                    1,
                                    0)
                                ) as %(sub_field_name)s
                                """
    }
    user_cols2 = """ "#lib"='Android' """
    tf_sql_obj = TrainFeatureSql(
        ctx={},
        gconf=gconf,
        need_events=need_events,
        need_event_attrs=need_event_attrs,
        feature_names=feature_names,
        sub_sql_fs=sub_sql_fs,
        user_cols2=user_cols2,
        addon_attrs=addon_attrs,  # 可选参数
        line_value_rd=0.5,  # 可选参数
        mode="iaa+iap"
    )
    sql = tf_sql_obj.get_sql()  # 这个sql就是拉取特征值的sql
    # 将sql复制到剪贴板
    import pyperclip
    pyperclip.copy(sql)
    return sql


if __name__ == "__main__":
    test()



'''
@des: tga获取训练特征值数据拉取的sql基类


准备工作: 
--------------------------------------
gconf的属性有这些:
########
   part_date_start  # 特征值取新增用户的开始日期, 例如"2022-05-20", 字符串类型
   part_date_end    # 特征值取新增用户的结束日期, 例如"2022-05-25", 字符串类型
   level_durs       # 特征取值段，由小到大排列, 例如 [43200, 43200*2], 这里是秒为单位，整型数组；多少时间段发生事件作为特征值
   label_durs       # 取多少边界算硬核用户标签，例如[int(86400*14), int(86400*17)]，整型数组；这里是取14到17天发生高粘度事件的作为正向用户
   tga_user_table   # tga用户表名, 例如 "v_user_7", 字符串
   tga_event_table  # tga事件表名, 例如 "v_event_7", 字符串 
########

如何使用(需要自己定义一个子类继承这个类， 代码如下): 
#######
from magic_number_train_tf_2.src.common.config import Config as GAConfig # 这里换成你自己的项目
from tfduck.tga.train_sql_yh import TrainFeatureSql

def call_method(self, ctx=None, **kwargs):
    """
    @des: 每个具体项目的训练特征数据拉取 
    """
    gconf = GAConfig.getInstance()  # 保证gconf有上面准备工作的属性
    """
    设置项目属性-----下面的都需要自己配置------
    """
    # 需要的事件
    need_events = (
        'new_device', 'new_session', 'act_level_path'
    )
    # 需要的属性
    need_event_attrs = (
        "#lib", "#country_code", "$part_event", "#user_id",  "sdk_session_time",
        "object_type", "act", "object_number", "act_object",
    )
    # 特征值名称--不要用字母a作为key---
    feature_names = {
        'b': '常规关卡通关',
        'c': '冒险关卡通关'
    }
    # 特征值sql---模板保持不变，变里面的内容即可---
    sub_sql_fs = {
        'b':            """
                        --%(real_des)s
                        sum(
                            if(
                            "$part_event"='act_level_path' and object_type='normal' and act='win' and rt_dur<%(level_dur)s,
                            1, 
                            0)
                        ) as %(sub_field_name)s
                        """,
        'c':            """
                        --%(real_des)s
                        sum(
                            if(
                            "$part_event"='act_level_path' and object_type='adventure' and act='win' and rt_dur<%(level_dur)s,
                            1, 
                            0)
                        ) as %(sub_field_name)s
                        """
    }    
    # 额外的event表属性和预查询的用户属性base_user_table_fields里面选，比如(可选参数)
    addon_attrs = ["#screen_width", "#screen_height"]        
    """
    创建调用实例---最好通过dict创建实例
    注意： 针对一个tga项目属于多个项目的打点情况,例如3dfish， 可以适时启用user_cols1,user_cols2或者event_cols1
    a.  user_cols1 参数为用户过滤条件(user表)
        如果有的话, 例如: user_cols1 = """ "afrawip__meida_source"='FaceBook' OR "afrawip__app_id"='123456' """
    b.  user_cols2 参数为用户过滤条件(event表的new_device事件)
        如果有的话, 例如: user_cols2 = """ "#lib"='Android' OR "#lib"='iOS' """
    c.  event_cols1 参数为事件过滤条件(event表的所有计算特征值的事件)
        如果有的话, 例如: event_cols1 = """ "#lib"='Android' OR "#lib"='iOS' """
    """         
    tf_sql_obj = TrainFeatureSql(
            ctx = ctx, 
            gconf = gconf, 
            need_events = need_events, 
            need_event_attrs = need_event_attrs, 
            feature_names = feature_names,
            sub_sql_fs = sub_sql_fs,
            addon_attrs = addon_attrs # 可选参数
    )  
    sql = tf_sql_obj.get_sql() # 这个sql就是拉取特征值的sql
#######
'''
from tfduck.tga.base_tga import BaseTga
from tfduck.common.defines import BMOBJ, Et
import arrow


class TrainFeatureSql(BaseTga):
    """
    @des:sql训练模板基类
    """

    def __init__(self,
                 ctx=None,
                 gconf=None,
                 need_events=None,
                 need_event_attrs=None,
                 feature_names=None,
                 sub_sql_fs=None,
                 user_cols1='1=1',
                 user_cols2='1=1',
                 event_cols1='1=1',
                 addon_attrs=[], 
                 **kwargs):
        """
        @des: 参数说明看上面的文档说明
        """
        self.gconf = gconf
        self.ctx = ctx
        # 基础属性
        self.label_col = "is_rd"
        self.base_event_table = 'all_need_event'
        self.base_user_table = 'a'
        self.base_feature_table = 'features_t'
        self.nav_neg_multi = 2  # 负正样本比例
        self.yh_value = 0.75  # 高粘度用户的粘度值
        # 项目属性
        self.need_events = need_events
        self.need_event_attrs = need_event_attrs
        self.feature_names = feature_names
        if self.base_user_table in self.feature_names:
            raise Et(2, f"attr name '{self.base_user_table}' cannt be used")
        feature_names_all = {
            self.base_user_table: '用户属性和标签',  # 此属性固定sql，不需要拼接
        }
        feature_names_all.update(self.feature_names)
        self.feature_names = feature_names_all
        self.sub_sql_fs = sub_sql_fs
        #
        self.user_cols1 = user_cols1
        self.user_cols2 = user_cols2
        self.event_cols1 = event_cols1
        self.addon_attrs = addon_attrs if isinstance(addon_attrs, list) else []
        # 其他属性
        for k, v in kwargs.items():
            setattr(self, k, v)

    def get_first_col_name(self):
        """
        @des: 获取第一个特征属性列
        """
        return f"{list(self.feature_names.keys())[1]}_0_v"

    def get_real_feature_names(self, mode='d'):
        """
        @des: 获取真正的特征值列表
        """
        real_features_names = {}
        sub_len = len(self.gconf.level_durs)
        for k, v in self.feature_names.items():
            if k != self.base_user_table:
                for i in range(sub_len):
                    if mode == 'h':
                        rv = f"{v}_{int(self.gconf.level_durs[i]/3600)}h"
                    else:
                        js = self.gconf.level_durs[i]/86400
                        if js == int(js):
                            js = int(js)
                        else:
                            js = round(js, 1)
                        rv = f"{v}_{js}d"
                    real_features_names[f"{k}_{i}_v"] = rv
        return real_features_names

    def get_threshold_sql(self, col_name):
        """
        @des: 根据col_name例如 g_1_v获取特征值计算的sql
        """
        first_col_name, second_col_index = col_name.split(
            "_")[0], col_name.split("_")[1]
        sub_field_name = "%s_%s_v" % (first_col_name, second_col_index)
        real_des = "%s %s" % (self.feature_names.get(
            first_col_name), second_col_index)
        level_dur = self.gconf.level_durs[int(second_col_index)]
        f_sub_sql = self.sub_sql_fs.get(first_col_name)
        if not f_sub_sql:
            raise Et(2, f"error first_col_name {first_col_name}")
        sub_sql = f_sub_sql % {
            "real_des": real_des, "level_dur": level_dur, 'sub_field_name': sub_field_name}
        return sub_sql

    def get_sql_config(self):
        """
        @des: 计算各种时间间隔和配置，统一从一个方法读取，后面方便改
        """
        gconf = self.gconf
        #
        part_date_start = gconf.part_date_start
        part_date_end = gconf.part_date_end
        days = (arrow.get(part_date_end)-arrow.get(part_date_start)).days
        # 取用户新增后的N天作为事件池数据来取特征值，比如7日预测14日，就取8-9比较合适，根据level_durs[-1]决定
        after_feature_day = int(gconf.level_durs[-1]/86400)+2  # 多取两天即可
        feature_date_end = arrow.get(part_date_end).shift(days=after_feature_day).format(
            "YYYY-MM-DD")  # 增加到8天的事件数据，因为现在有7日预测30日
        label_durs = gconf.label_durs
        label_date_start = arrow.get(part_date_start).shift(
            days=int(label_durs[0]/86400)-2).format("YYYY-MM-DD")
        label_date_end = arrow.get(part_date_end).shift(
            days=int(label_durs[1]/86400)+2).format("YYYY-MM-DD")
        #
        new_device_start = arrow.get(part_date_start).format("YYYY-MM-DD")
        new_device_end = arrow.get(part_date_end).shift(
            days=1).format("YYYY-MM-DD")
        return {
            "days": days,
            'part_date_start': part_date_start,
            'part_date_end': part_date_end,
            'feature_date_end': feature_date_end,
            'label_date_start': label_date_start,
            'label_date_end': label_date_end,
            'new_device_start':  new_device_start,
            'new_device_end':  new_device_end
        }

    def get_event_sql(self):
        """
        @des: 获取N日留存用户的事件
        """
        gconf = self.gconf
        comm_cc = self.get_sql_config()
        #
        sql = f"""
    -- des: add by yuanxiao for machine learn train
    with new_user as (
                -- 获取指定日期的注册用户 连接 这些用户在7-10天后的触发的事件
                SELECT
                    a1.user_register_time as user_register_time,
                    a1."#user_id" as user_user_id,
                    b1."#user_id" as event_user_id,
                    -- b1."event_time_utc" as pd_event_time_utc,
                    floor(to_unixtime(b1.event_time_utc))-floor(to_unixtime(a1.user_register_time)) as pd_rt_dur
                FROM
                (
                    -- 获取指定日期的注册用户,以防万一，使用窗口函数去个重
                    SELECT
                        *
                    FROM
                    (
                        SELECT
                            *,
                            row_number()  OVER  (PARTITION  BY  "#user_id" ORDER BY user_register_time) AS  row_no
                        FROM
                        (
                            SELECT
                                b.event_time_utc as user_register_time, a."#user_id"
                            FROM
                            (
                                SELECT
                                    "#user_id"
                                FROM
                                    {gconf.tga_user_table}
                                WHERE
                                    {self.user_cols1}
                            ) a
                            INNER JOIN
                            (
                                SELECT
                                    *
                                FROM (
                                    SELECT
                                        "#user_id",
                                        {self.tran_dt_by_zone("#event_time", "#zone_offset", 0)} as event_time_utc,
                                        "$part_event"
                                    FROM
                                        {gconf.tga_event_table}
                                    WHERE
                                        "$part_event" = 'new_device'
                                        AND "$part_date" >= '{arrow.get(comm_cc["new_device_start"]).shift(days=-2).format("YYYY-MM-DD")}'
                                        AND "$part_date" <= '{arrow.get(comm_cc["new_device_end"]).shift(days=2).format("YYYY-MM-DD")}'
                                        AND {self.user_cols2}
                                )
                                WHERE
                                    event_time_utc >= timestamp '{comm_cc["new_device_start"]}'
                                    AND event_time_utc < timestamp '{comm_cc["new_device_end"]}'
                            ) b
                            ON a."#user_id" = b."#user_id"
                        )
                    )
                    WHERE row_no=1
                ) a1
                LEFT JOIN
                (
                    -- 获取指定日期的label_durs的高粘度条件
                    SELECT
                        *
                    FROM (
                        SELECT
                            "#user_id", 
                            sdk_retention_day,
                            sdk_play_day,
                            {self.tran_dt_by_zone("#event_time", "#zone_offset", 0)} as event_time_utc
                        FROM
                            {gconf.tga_event_table}
                        WHERE
                            "$part_event" = 'new_session'
                            AND "$part_date" >= '{arrow.get(comm_cc["label_date_start"]).shift(days=-2).format("YYYY-MM-DD")}'
                            AND "$part_date" <= '{arrow.get(comm_cc["label_date_end"]).shift(days=2).format("YYYY-MM-DD")}'
                    )
                    WHERE
                        event_time_utc >= timestamp '{comm_cc["label_date_start"]}'
                        AND event_time_utc <= timestamp '{comm_cc["label_date_end"]}'  
                        -- 加上高粘度条件判断 
                        AND sdk_retention_day>={int(gconf.label_durs[0]/86400)}
                        AND sdk_retention_day<={int(gconf.label_durs[1]/86400)}
                        AND (cast(sdk_play_day as double)/(cast(sdk_retention_day as double)+1))>{self.yh_value}
                ) b1
                ON a1."#user_id" = b1."#user_id"
    )
    , user_label_table as (
        -- 获取指定日期的注册用户的训练标签
        SELECT
            CASE WHEN a3.event_count>0 THEN '1' ELSE '0' END as is_rd,
            a3.user_user_id,
            a3.user_register_time,
            -- 计算固定随机值，打乱顺序
            (abs(from_ieee754_64(xxhash64(cast(cast(user_user_id as varchar) as varbinary)))) % 100) / 100. as tt_stable_rand 
        FROM
        (
            -- 获取指定日期的注册用户 连接 这些用户在7-10天后的触发的事件 的 数量
            SELECT
                a2.user_user_id as user_user_id,
                a2.user_register_time as user_register_time,
                SUM(CASE WHEN a2.event_user_id IS NULL THEN 0 ELSE 1 END) AS event_count
            FROM
            (
                new_user
            ) a2
            GROUP BY a2.user_user_id, a2.user_register_time
        ) a3
    )
    , nav_table as (
        select  
            *
        from 
            user_label_table  
        where 
            is_rd='1'
        -- 固定排序，防止找不到用户, 不适用于超大数据量
        order by tt_stable_rand
        limit {6000* comm_cc['days']}
    )
    , neg_table as (
        select  
            *
        from 
            user_label_table  
        where 
            is_rd='0'
        -- 固定排序，防止找不到用户, 不适用于超大数据量
        order by tt_stable_rand
        limit {12000* comm_cc['days']}
    )
    , union_all as (
        -- 保持正负样本固定比例1:2
        select
            *
        from (
            (
                select is_rd,user_register_time, user_user_id  from nav_table
            )
            UNION ALL
            (
                select
                    is_rd,user_register_time, user_user_id
                from (
                    SELECT *, row_number() OVER (
                        PARTITION BY is_rd 
                        ORDER BY tt_stable_rand
                    ) AS kere_nopoa_end_0
                    FROM neg_table
                ) where kere_nopoa_end_0 < (select count(1)*{self.nav_neg_multi} from nav_table)
            ) 
        ) st
    )
    , all_need_event as (
        SELECT
            *
        FROM (
            -- 获取正负标签样本需要的事件
            select 
                a5.is_rd,
                a5.user_register_time,
                a5.user_user_id,
                floor(to_unixtime(b5.event_time_utc))-floor(to_unixtime(a5.user_register_time)) as rt_dur,
                c5."afrawip__meida_source" as "afrawip__meida_source_ikfdssausercommonend",
                c5."#account_id" as "#account_id_ikfdssausercommonend",
                c5."#distinct_id" as "#distinct_id_ikfdssausercommonend",
                c5."afrawip__campaign" as "afrawip__campaign_ikfdssausercommonend",
                b5.*,
                b5.event_time_utc as "#event_time"
            from 
                union_all a5
            INNER JOIN
            (
                SELECT
                    *
                FROM (
                    SELECT
                        -- *
                        {','.join(['"%s"'%x for x in self.need_event_attrs])},
                        {self.tran_dt_by_zone("#event_time", "#zone_offset", 0)} as event_time_utc
                    FROM
                        {gconf.tga_event_table}
                    WHERE
                        "$part_date" >= '{arrow.get(comm_cc["part_date_start"]).shift(days=-2).format("YYYY-MM-DD")}'
                        AND "$part_date" <= '{arrow.get(comm_cc["feature_date_end"]).shift(days=2).format("YYYY-MM-DD")}'
                        AND "$part_event" in ({','.join(["'%s'"%x for x in self.need_events])})
                        AND {self.event_cols1}      
                )
                WHERE
                    event_time_utc >= timestamp '{comm_cc["part_date_start"]}'
                    AND event_time_utc <= timestamp '{comm_cc["feature_date_end"]}'    
            )
            as b5
            ON a5.user_user_id = b5."#user_id"
            INNER JOIN
            (
                SELECT
                    "#user_id","afrawip__meida_source","#account_id","#distinct_id","afrawip__campaign"
                FROM
                    {gconf.tga_user_table}
            )
            as c5
            ON a5.user_user_id = c5."#user_id"
        )  a6
        WHERE
        rt_dur>=0 AND rt_dur<={gconf.level_durs[-1]}  -- N小时内的事件
    ) 
    """
        return sql

    def get_sub_sql_i(self, sub_sql_f, col_name, col_des, base_event_table='all_need_event', base_user_table='a'):
        """
        @des: 内部调用
        """
        gconf = self.gconf
        sub_field_f = "%(col)s_%(dur_i)s"
        real_des_f = "%(des)s %(dur_i)s"
        sub_field_names = []
        sub_sqls = []
        for _dur_i, level_dur in enumerate(gconf.level_durs):
            real_des = real_des_f % {'des': col_des, 'dur_i': _dur_i}
            sub_field_name = sub_field_f % {'col': col_name, 'dur_i': _dur_i}
            sub_field_names.append(sub_field_name)
            # sub_sql_f格式如下
            # sub_sql_f = """
            #     --%(real_des)s
            #     sum(
            #         if(
            #             "$part_event"=='act_level_path' and object_type='normal' and act='win' and rt_dur<%(level_dur)s,
            #             1,
            #             0)
            #     ) as %(sub_field_name)s
            # """
            sub_sql = sub_sql_f % {
                'real_des': real_des,
                'level_dur': level_dur,
                'sub_field_name': sub_field_name
            }
            sub_sqls.append(sub_sql)
        return sub_field_names, sub_sqls

    def get_feature_sql(self):
        """
        @des: 提取特征值的sql
        """
        gconf = self.gconf
        """
        构建level_durs的sql
        """
        base_event_table = self.base_event_table
        base_user_table = self.base_user_table
        base_feature_table = self.base_feature_table
        feature_names = self.feature_names
        comm_cc = self.get_sql_config()
        #
        sub_sqls = []
        sub_field_names = []
        for col_name, col_des in feature_names.items():
            sub_sql_f = ""
            if col_name == 'a':
                continue
            else:
                sub_sql_f = self.sub_sql_fs.get(col_name)
                if not sub_sql_f:
                    raise Et(2, f"error col_name: {col_name} ")
            # 拼接结果
            if sub_sql_f:
                _sub_field_names, _sub_sqls = self.get_sub_sql_i(
                    sub_sql_f, col_name, col_des)
                sub_sqls.extend(_sub_sqls)
                sub_field_names.extend(_sub_field_names)

        """
        构建select选项
        """
        # 带表名的字段 表名.field
        base_user_table_fields = [
            "is_rd", "user_user_id", "afrawip__meida_source_ikfdssausercommonend", "#lib", "#country_code"]
        for addon_attr in self.addon_attrs:
            _addon_attr = addon_attr.strip()
            if _addon_attr not in base_user_table_fields:
                base_user_table_fields.append(_addon_attr)
        #
        sub_selects = ",".join(
            [f"{base_feature_table}.{item} as {item}_v" for item in sub_field_names])
        sub_selects = f"""{",".join(['%s."%s"'%(base_user_table, ptf) for ptf in base_user_table_fields])},{sub_selects}"""
        # 不带表名的字段
        # sub_selects_unique = ",".join(
        #     [f'"{item}_v"' for item in sub_field_names])
        # sub_selects_unique = f"""{",".join(['"%s"'%ptf for ptf in base_user_table_fields])},{sub_selects_unique}"""
        """
        构建sub_sqls
        """
        sub_join_sqls = ",".join(sub_sqls)
        """
        拼接最后的sql
        """
        sql = f"""
    , user_tzz as (
        select 
            {sub_selects}
        from (
                --用户属性和标签
                select 
                    {",".join(['"%s"'%ptf for ptf in base_user_table_fields])}
                from
                (
                    select
                        {",".join(['"%s"'%ptf for ptf in base_user_table_fields])},
                        row_number()  OVER  (PARTITION  BY  user_user_id ORDER  BY  "#event_time" asc) AS  row_no 
                    from 
                        {base_event_table} 
                )
                where
                    row_no=1
        ) {base_user_table} 
        left join
        (   --- 特征值计算
            select
                user_user_id,
                {sub_join_sqls}
            from
                {base_event_table}
            group by user_user_id
        ) {base_feature_table}
        on {base_user_table}.user_user_id={base_feature_table}.user_user_id
    )
    -- with结束没有逗号, 过滤左连接没有特征值的行，调试的时候取消where条件
    select * from user_tzz where {self.get_first_col_name()} is not NULL
    """
        return sql

    def get_sql(self):
        # print("---------------------------------")
        event_sql = self.get_event_sql()
        feature_sql = self.get_feature_sql()
        sql = f"""
    {event_sql}
    {feature_sql} 
    """
        return sql
